# -*- coding: utf-8 -*-
from collections import defaultdict
import datetime
import logging
import tempfile
import zipfile
from zipfile import ZIP_DEFLATED
from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.db import connection
from django.http import HttpResponse
from django.shortcuts import render
from django.views.decorators.http import require_POST
from dates2utils.dates2export import (
    xml_teilnehmerDatenImport,
    update_by_timeinterval
)
from reportserver.lib.io.requests import (
    check_date,
    convert_german_date_to_iso
)


logger = logging.getLogger(__name__)


def dict_cursor(cursor):
    description = [x[0] for x in cursor.description]
    for row in cursor:
        yield dict(zip(description, row))


def extract_program_name(beneficiary_id):
    """Extract the program name from beneficiary_id.

    By definition the first section, seperated by '.' (dot).
    """
    return beneficiary_id.split('.')[0]


def get_xml_filename(beneficiary_and_rc, time_str):
    """Build filename from normalized beneficiary id, region category (EC only)
    and time_str."""
    return u"%s_%s.xml" % (beneficiary_and_rc, time_str)


def update_indicator_name(row, indicators):
    """It is unclear why the indicator name update is necessary.

    This should be explained.
    """
    r = {}
    for key in row:
        if key.upper() in indicators:
            r[key.upper()] = row[key]
        elif key.upper().replace("_", ".") in indicators:
            r[key.upper().replace("_", ".")] = row[key]
        else:
            r[key] = row[key]
    return r


def personal_id_to_instanceid(value, separator="-", merger="."):
    split_value = value.split(separator)
    return merger.join(split_value[0:2])


def create_instanceid_rows_dict(rows, program):
    result_dict = defaultdict(list)
    for row in rows:
        # the instance_id is a shorted version.
        instance_id = personal_id_to_instanceid(row['personal_id'])
        # EC have different handling of the region_category.
        if program == "EC":
            result_dict[(instance_id,
                         str(row["regioncategory_id"]))].append(row)
        else:
            result_dict[(instance_id, "*")].append(row)
    return result_dict


def get_GUID(name, session):
    return session.get('dates2_lut')[stringify_pair(*name)]


def stringify_pair(p1, p2):
    return "%s;%s" % (p1, p2)


def get_full_beneficiary(string1, beneficiaries):
    # calc the year number of the beneficiary_id
    for beneficiary in beneficiaries:
        instance_id_with_underscores = beneficiary[0]
        if instance_id_with_underscores.startswith(string1):
            return instance_id_with_underscores
    return string1


def get_name_from(pair, beneficiaries):
    # handling the special EC case (in the conditional branch)
    name = get_full_beneficiary(pair[0], beneficiaries)
    if not pair[1] == "*":
        name += ("_" + str(pair[1]))
    return name


def handle_beneficiary(beneficiary, cursor, start, end, time_str, zip1,
                       session, beneficiaries):
    cursor.execute("SELECT * FROM eval_virtual_participations"
                   " WHERE beneficiary_id=%s;", [beneficiary[0]])
    rows = [x for x in dict_cursor(cursor)]
    rows = update_by_timeinterval(
        rows, start, end,
        {'RI': settings.DATES2_RI_INDICATORS})
    out = []
    oi_indicators = settings.DATES2_OI_INDICATORS
    ri_indicators = settings.DATES2_RI_INDICATORS
    indicators = oi_indicators + ri_indicators
    rows_updated = [update_indicator_name(row, indicators)
                    for row in rows]
    # Group the rows by organisation id and region category:
    # Build a list of rows for each group to handle the
    # EC case and organisation changes.
    rows_dict = create_instanceid_rows_dict(
        rows_updated,
        extract_program_name(beneficiary[0]))

    for pair in rows_dict:
        out = []
        xml_teilnehmerDatenImport(out, get_GUID(pair, session),
                                  rows_dict[pair],
                                  oi_indicators, ri_indicators)
        output = u"\n".join(out)
        zip1.writestr(get_xml_filename(get_name_from(pair, beneficiaries),
                                       time_str),
                      output)


def get_date_from(request, name):
    return datetime.datetime.strptime(check_date(convert_german_date_to_iso(
        request.POST.get(name))), '%Y-%m-%d').date()


def handle_insert_lists(cursor, start, end, time_str, session, beneficiaries,
                        main_zip):
    # handle normal xml part
    with tempfile.TemporaryFile() as ziptmpfile2:
        with zipfile.ZipFile(ziptmpfile2, "w", ZIP_DEFLATED) as zip2:
            # for every b: 1. get rows 2. xml them
            for beneficiary in beneficiaries:
                handle_beneficiary(beneficiary, cursor, start, end,
                                   time_str, zip2, session,
                                   beneficiaries)
        ziptmpfile2.seek(0)
        progname = get_progname_by_beneficiaries(beneficiaries)
        main_zip.writestr(create_normal_zip_arcname(progname, time_str),
                          ziptmpfile2.read())


def create_normal_zip_arcname(program, time_str):
    """create the arcname for the normal xml zip."""
    return "Teilnehmerlisten-%s_%s.zip" % (program, time_str)


# code for generation of the remove lists
def create_rm_list_xml_arcname(beneficiary, number, time_str):
    """create the arcname for a rm list xml."""
    return "%s-%d_%s.xml" % (beneficiary, number, time_str)


def create_rm_lists_zip_arcname(program, number, time_str, region_category):
    """create the arcname for a rm lists zip."""
    if region_category == "*":
        return "Loeschlisten-%s_%s-%d.zip" % (program, time_str, number)
    return "Loeschlisten-%s-%s_%s-%d.zip" % (program, region_category,
                                             time_str, number)


def chunk_2(seq, n):
    """Split sequence seq in max n large chunks."""
    i = 0
    while seq[i:]:
        yield seq[i:i + n]
        i += n


def generate_rm_list_xml(entry, session, rc="*"):
    out = []
    instanceid = personal_id_to_instanceid(entry[1])
    xml_teilnehmerDatenImport(out, get_GUID((instanceid, rc), session),
                              [], [], [],
                              loeschliste=[{"participant_uuid": entry[0],
                                            "personal_id": entry[1]}])
    output = u"\n".join(out)
    return output


MAX_FILES_IN_ZIP = 65000


def handle_rm_lists(cursor, time_str, main_zip, beneficiaries, session):
    """Handle the create of the rm lists (XML-Löschlisten), write the Zips
    and insert them into the main_zip."""
    cursor.execute("SELECT uuid, personal_id FROM participant_ids;")
    rows = [row for row in cursor.fetchall()]
    region_categories = ["*"]
    if settings.PROGRAM == "EC":
        region_categories = ["1", "2", "3", "4"]
    for nc_rc, region_category in enumerate(region_categories):
        chunks = chunk_2(rows, MAX_FILES_IN_ZIP)
        nr_entries = 1
        nr_chunks = 1
        for chunk in chunks:
            with tempfile.TemporaryFile() as ziptmpfile:
                ZDE = ZIP_DEFLATED
                with zipfile.ZipFile(ziptmpfile, "w", ZDE) as secondary_zip:
                    for entry in chunk:
                        xml = generate_rm_list_xml(entry, session,
                                                   rc=region_category)
                        instance_id_full = get_full_beneficiary(
                            personal_id_to_instanceid(entry[1]), beneficiaries)
                        if settings.PROGRAM == "EC":
                            instance_id_full += (u"-%s" % region_category)
                        arcname = create_rm_list_xml_arcname(
                            instance_id_full, nr_entries, time_str)
                        secondary_zip.writestr(arcname, xml)
                        nr_entries += 1
                ziptmpfile.seek(0)
                zip_str = ziptmpfile.read()
                progname = get_progname_by_beneficiaries(beneficiaries)
                zip_arcname = create_rm_lists_zip_arcname(progname,
                                                          nr_chunks,
                                                          time_str,
                                                          region_category)
                main_zip.writestr(zip_arcname, zip_str)
            nr_chunks += 1


# end rm list stuff

def get_progname_by_beneficiaries(beneficiaries):
    if beneficiaries:
        return extract_program_name(beneficiaries[0][0])
    else:
        return settings.PROGRAM


@login_required
@require_POST
def export_dates2(request):
    time_str = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    start = get_date_from(request, 'start_date')
    end = get_date_from(request, 'end_date')

    # Load list of beneficiaries (e.g.): id (PRJ.1234.15), label (PRJ Muster)
    beneficiaries = []
    with connection.cursor() as cursor:
        cursor.execute("SELECT id FROM beneficiary ORDER BY id;")
        beneficiaries = [row for row in cursor.fetchall()]

        with tempfile.TemporaryFile() as ziptmpfile:
            with zipfile.ZipFile(ziptmpfile, "w", ZIP_DEFLATED) as zip1:

                handle_rm_lists(cursor, time_str, zip1, beneficiaries,
                                request.session)
                handle_insert_lists(cursor, start, end, time_str,
                                    request.session, beneficiaries, zip1)
            ziptmpfile.seek(0)
            zip_str = ziptmpfile.read()

    program_name = get_progname_by_beneficiaries(beneficiaries)
    response = HttpResponse(zip_str, content_type="application/zip")
    response['Content-Disposition'] = ('attachment; filename="%s_%s.zip"' %
                                       (program_name, time_str))
    logger.info("'%s' requested a dates2export.", request.user.username)
    return response


@login_required
def overview(request):
    return render(request, 'dates2export.html', {})
